@@ -5,6 +5,8 @@ Some of the utils used by salt
 from __future__ import absolute_import
 
 # Import python libs
+import copy
+import collections
 import datetime
 import distutils.version  # pylint: disable=E0611
 import fnmatch
@@ -19,9 +21,9 @@ import random
 import re
 import shlex
 import shutil
-import string
 import socket
 import stat
+import string
 import subprocess
 import sys
 import tempfile
@@ -31,6 +33,13 @@ import warnings
 import yaml
 from calendar import month_abbr as months
 
+# Try to load pwd, fallback to getpass if unsuccessful
+try:
+    import pwd
+except ImportError:
+    import getpass
+    pwd = None
+
 try:
     import timelib
     HAS_TIMELIB = True
@@ -63,6 +72,20 @@ except ImportError:
     # Running as purely local
     pass
 
+try:
+    import grp
+    HAS_GRP = True
+except ImportError:
+    # grp is not available on windows
+    HAS_GRP = False
+
+try:
+    import pwd
+    HAS_PWD = True
+except ImportError:
+    # pwd is not available on windows
+    HAS_PWD = False
+
 # Import salt libs
 import salt._compat
 import salt.log
@@ -72,7 +95,7 @@ import salt.version
 from salt._compat import string_types
 from salt.utils.decorators import memoize as real_memoize
 from salt.exceptions import (
-    SaltClientError, CommandNotFoundError, SaltSystemExit
+    SaltClientError, CommandNotFoundError, SaltSystemExit, SaltInvocationError
 )
 
 
@@ -98,10 +121,11 @@ DEFAULT_COLOR = '\033[00m'
 RED_BOLD = '\033[01;31m'
 ENDC = '\033[0m'
 
-#KWARG_REGEX = re.compile(r'^([^\d\W][\w-]*)=(?!=)(.*)$', re.UNICODE)  # python 3
-KWARG_REGEX = re.compile(r'^([^\d\W][\w-]*)=(?!=)(.*)$')
+#KWARG_REGEX = re.compile(r'^([^\d\W][\w.-]*)=(?!=)(.*)$', re.UNICODE)  # python 3
+KWARG_REGEX = re.compile(r'^([^\d\W][\w.-]*)=(?!=)(.*)$')
 
 log = logging.getLogger(__name__)
+_empty = object()
 
 
 def get_function_argspec(func):
@@ -176,6 +200,11 @@ def get_colors(use=True):
     if not use:
         for color in colors:
             colors[color] = ''
+    if isinstance(use, str):
+        # Try to set all of the colors to the passed color
+        if use in colors:
+            for color in colors:
+                colors[color] = colors[use]
 
     return colors
 
@@ -212,9 +241,23 @@ def get_context(template, line, num_lines=5, marker=None):
     if marker:
         buf[error_line_in_context] += marker
 
+    # warning: jinja content may contain unicode strings
+    # instead of utf-8.
+    buf = [i.encode('UTF-8') if isinstance(i, unicode) else i for i in buf]
+
     return '---\n{0}\n---'.format('\n'.join(buf))
 
 
+def get_user():
+    '''
+    Get the current user
+    '''
+    if pwd is not None:
+        return pwd.getpwuid(os.geteuid()).pw_name
+    else:
+        return getpass.getuser()
+
+
 def daemonize(redirect_out=True):
     '''
     Daemonize a process
@@ -232,6 +275,7 @@ def daemonize(redirect_out=True):
 
     # decouple from parent environment
     os.chdir('/')
+    # noinspection PyArgumentList
     os.setsid()
     os.umask(18)
 
@@ -253,7 +297,7 @@ def daemonize(redirect_out=True):
     # not cleanly redirected and the parent process dies when the
     # multiprocessing process attempts to access stdout or err.
     if redirect_out:
-        dev_null = open('/dev/null', 'w')
+        dev_null = open('/dev/null', 'r+')
         os.dup2(dev_null.fileno(), sys.stdin.fileno())
         os.dup2(dev_null.fileno(), sys.stdout.fileno())
         os.dup2(dev_null.fileno(), sys.stderr.fileno())
@@ -325,7 +369,21 @@ def which(exe=None):
                     continue
             return False
 
-        for path in search_path.split(os.pathsep):
+        search_path = search_path.split(os.pathsep)
+        if not is_windows():
+            # Add any dirs in the default_path which are not in search_path. If
+            # there was no PATH variable found in os.environ, then this will be
+            # a no-op. This ensures that all dirs in the default_path are
+            # searched, which lets salt.utils.which() work well when invoked by
+            # salt-call running from cron (which, depending on platform, may
+            # have a severely limited PATH).
+            search_path.extend(
+                [
+                    x for x in default_path.split(os.pathsep)
+                    if x not in search_path
+                ]
+            )
+        for path in search_path:
             full_path = os.path.join(path, exe)
             if os.access(full_path, os.X_OK):
                 return full_path
@@ -344,7 +402,9 @@ def which(exe=None):
             )
         )
     else:
-        log.trace('No executable was passed to be searched by which')
+        log.error(
+            'No executable was passed to be searched by salt.utils.which()'
+        )
     return None
 
 
@@ -352,7 +412,7 @@ def which_bin(exes):
     '''
     Scan over some possible executables and return the first one that is found
     '''
-    if not isinstance(exes, (list, tuple)):
+    if not isinstance(exes, collections.Iterable):
         return None
     for exe in exes:
         path = which(exe)
@@ -454,6 +514,9 @@ def dns_check(addr, safe=False, ipv6=False):
                     break
             if not addr:
                 error = True
+    except TypeError:
+        err = ('Attempt to resolve address failed. Invalid or unresolveable address')
+        raise SaltSystemExit(code=42, msg=err)
     except socket.error:
         error = True
 
@@ -550,7 +613,7 @@ def jid_load(jid, cachedir, sum_type, serial='msgpack'):
     if not os.path.isfile(load_fn):
         return {}
     serial = salt.payload.Serial(serial)
-    with fopen(load_fn) as fp_:
+    with fopen(load_fn, 'rb') as fp_:
         return serial.load(fp_)
 
 
@@ -567,7 +630,6 @@ def is_jid(jid):
         return True
     except ValueError:
         return False
-    return False
 
 
 def check_or_die(command):
@@ -596,7 +658,7 @@ def copyfile(source, dest, backup_mode='', cachedir=''):
         )
     if not os.path.isdir(os.path.dirname(dest)):
         raise IOError(
-            '[Errno 2] No such file or directory: {0}'.format(source)
+            '[Errno 2] No such file or directory: {0}'.format(dest)
         )
     bname = os.path.basename(dest)
     dname = os.path.dirname(os.path.abspath(dest))
@@ -611,7 +673,18 @@ def copyfile(source, dest, backup_mode='', cachedir=''):
     if backup_mode == 'master' or backup_mode == 'both' and bkroot:
         # TODO, backup to master
         pass
+    # Get current file stats to they can be replicated after the new file is
+    # moved to the destination path.
+    fstat = None
+    if not salt.utils.is_windows():
+        try:
+            fstat = os.stat(dest)
+        except OSError:
+            pass
     shutil.move(tgt, dest)
+    if fstat is not None:
+        os.chown(dest, fstat.st_uid, fstat.st_gid)
+        os.chmod(dest, fstat.st_mode)
     # If SELINUX is available run a restorecon on the file
     rcon = which('restorecon')
     if rcon:
@@ -633,7 +706,7 @@ def backup_minion(path, bkroot):
     dname, bname = os.path.split(path)
     fstat = os.stat(path)
     msecs = str(int(time.time() * 1000000))[-6:]
-    stamp = time.asctime().replace(' ', '_')
+    stamp = time.strftime('%a_%b_%d_%H:%M:%S_%Y')
     stamp = '{0}{1}_{2}'.format(stamp[:-4], msecs, stamp[-4:])
     bkpath = os.path.join(bkroot,
                           dname[1:],
@@ -734,47 +807,140 @@ def build_whitespace_split_regex(text):
 
 
 def build_whitepace_splited_regex(text):
-    warnings.warn("The build_whitepace_splited_regex function is deprecated,"
-                  " please use build_whitespace_split_regex instead.",
+    warnings.warn('The build_whitepace_splited_regex function is deprecated,'
+                  ' please use build_whitespace_split_regex instead.',
                   DeprecationWarning)
     build_whitespace_split_regex(text)
 
 
-def format_call(fun, data):
+def format_call(fun,
+                data,
+                initial_ret=None,
+                expected_extra_kws=()):
     '''
-    Pass in a function and a dict containing arguments to the function.
+    Build the required arguments and keyword arguments required for the passed
+    function.
 
-    A dict with the keys args and kwargs is returned
+    :param fun: The function to get the argspec from
+    :param data: A dictionary containing the required data to build the
+                 arguments and keyword arguments.
+    :param initial_ret: The initial return data pre-populated as dictionary or
+                        None
+    :param expected_extra_kws: Any expected extra keyword argument names which
+                               should not trigger a :ref:`SaltInvocationError`
+    :returns: A dictionary with the function required arguments and keyword
+              arguments.
     '''
-    ret = {}
+    ret = initial_ret is not None and initial_ret or {}
+
     ret['args'] = []
-    aspec = get_function_argspec(fun)
-    if aspec.keywords:
-        # This state accepts kwargs
-        ret['kwargs'] = {}
-        for key in data:
-            # Passing kwargs the conflict with args == traceback
-            if key in aspec.args:
-                continue
-            ret['kwargs'][key] = data[key]
+    ret['kwargs'] = {}
 
-    try:
-        kwargs = dict(zip(aspec.args[::-1], aspec.defaults[::-1]))
-    except TypeError:
-        # aspec.defaults is None
-        kwargs = {}
-    for arg in kwargs:
-        if arg in data:
-            kwargs[arg] = data[arg]
-    for arg in aspec.args:
-        if arg in kwargs:
-            ret['args'].append(kwargs[arg])
+    aspec = get_function_argspec(fun)
+
+    args, kwargs = arg_lookup(fun).values()
+
+    # Since we WILL be changing the data dictionary, let's change a copy of it
+    data = data.copy()
+
+    missing_args = []
+
+    for key in kwargs.keys():
+        try:
+            kwargs[key] = data.pop(key)
+        except KeyError:
+            # Let's leave the default value in place
+            pass
+
+    while args:
+        arg = args.pop(0)
+        try:
+            ret['args'].append(data.pop(arg))
+        except KeyError:
+            missing_args.append(arg)
+
+    if missing_args:
+        used_args_count = len(ret['args']) + len(args)
+        args_count = used_args_count + len(missing_args)
+        raise SaltInvocationError(
+            '{0} takes at least {1} argument{2} ({3} given)'.format(
+                fun.__name__,
+                args_count,
+                args_count > 1 and 's' or '',
+                used_args_count
+            )
+        )
+
+    ret['kwargs'].update(kwargs)
+
+    if aspec.keywords:
+        # The function accepts **kwargs, any non expected extra keyword
+        # arguments will made available.
+        for key, value in data.iteritems():
+            if key in expected_extra_kws:
+                continue
+            ret['kwargs'][key] = value
+
+        # No need to check for extra keyword arguments since they are all
+        # **kwargs now. Return
+        return ret
+
+    # Did not return yet? Lets gather any remaining and unexpected keyword
+    # arguments
+    extra = {}
+    for key, value in data.iteritems():
+        if key in expected_extra_kws:
+            continue
+        extra[key] = copy.deepcopy(value)
+
+    # We'll be showing errors to the users until Salt Lithium comes out, after
+    # which, errors will be raised instead.
+    warn_until(
+        'Lithium',
+        'It\'s time to start raising `SaltInvocationError` instead of '
+        'returning warnings',
+        # Let's not show the deprecation warning on the console, there's no
+        # need.
+        _dont_call_warnings=True
+    )
+
+    if extra:
+        # Found unexpected keyword arguments, raise an error to the user
+        if len(extra) == 1:
+            msg = '{0[0]!r} is an invalid keyword argument for {1!r}'.format(
+                extra.keys(),
+                ret.get(
+                    # In case this is being called for a state module
+                    'full',
+                    # Not a state module, build the name
+                    '{0}.{1}'.format(fun.__module__, fun.__name__)
+                )
+            )
         else:
-            try:
-                ret['args'].append(data[arg])
-            except KeyError:
-                # Bad arg match, can safely proceed
-                pass
+            msg = '{0} and {1!r} are invalid keyword arguments for {2}'.format(
+                ', '.join(['{0!r}'.format(e) for e in extra][:-1]),
+                extra.keys()[-1],
+                ret.get(
+                    # In case this is being called for a state module
+                    'full',
+                    # Not a state module, build the name
+                    '{0}.{1}'.format(fun.__module__, fun.__name__)
+                )
+            )
+
+        # Return a warning to the user explaining what's going on
+        ret.setdefault('warnings', []).append(
+            '{0}. If you were trying to pass additional data to be used '
+            'in a template context, please populate \'context\' with '
+            '\'key: value\' pairs. Your approach will work until Salt Lithium '
+            'is out.{1}'.format(
+                msg,
+                '' if 'full' not in ret else ' Please update your state files.'
+            )
+        )
+
+        # Lets pack the current extra kwargs as template context
+        ret.setdefault('context', {}).update(extra)
     return ret
 
 
@@ -978,7 +1144,7 @@ def traverse_dict(data, key, default, delim=':'):
     Traverse a dict using a colon-delimited (or otherwise delimited, using
     the "delim" param) target string. The target 'foo:bar:baz' will return
     data['foo']['bar']['baz'] if this value exists, and will otherwise
-    return an empty dict.
+    return the dict in the default argument.
     '''
     try:
         for each in key.split(delim):
@@ -1030,9 +1196,17 @@ def is_windows():
 @real_memoize
 def is_linux():
     '''
-    Simple function to return if a host is Linux or not
+    Simple function to return if a host is Linux or not.
+    Note for a proxy minion, we need to return something else
     '''
-    return sys.platform.startswith('linux')
+    import __main__ as main
+    # This is a hack.  If a proxy minion is started by other
+    # means, e.g. a custom script that creates the minion objects
+    # then this will fail.
+    if 'salt-proxy-minion' in main.__file__:
+        return False
+    else:
+        return sys.platform.startswith('linux')
 
 
 @real_memoize
@@ -1066,31 +1240,30 @@ def check_state_result(running):
     '''
     if not isinstance(running, dict):
         return False
+
     if not running:
         return False
-    for host in running:
-        if not isinstance(running[host], dict):
-            return False
 
-        if host.find('_|-') == 4:
-            # This is a single ret, no host associated
-            rets = running[host]
-        else:
-            rets = running[host].values()
-
-        if isinstance(rets, dict) and 'result' in rets:
-            if rets['result'] is False:
-                return False
-            return True
-
-        for ret in rets:
-            if not isinstance(ret, dict):
-                return False
-            if 'result' not in ret:
-                return False
-            if ret['result'] is False:
-                return False
-    return True
+    ret = True
+    for state_result in running.itervalues():
+        if not isinstance(state_result, dict):
+            # return false when hosts return a list instead of a dict
+            ret = False
+        if ret:
+            result = state_result.get('result', _empty)
+            if result is False:
+                ret = False
+            # only override return value if we are not already failed
+            elif (
+                result is _empty
+                and isinstance(state_result, dict)
+                and ret
+            ):
+                ret = check_state_result(state_result)
+        # return as soon as we got a failure
+        if not ret:
+            break
+    return ret
 
 
 def test_mode(**kwargs):
@@ -1397,21 +1570,21 @@ def date_format(date=None, format="%Y-%m-%d"):
     >>> import datetime
     >>> src = datetime.datetime(2002, 12, 25, 12, 00, 00, 00)
     >>> date_format(src)
-    'Dec 25, 2002'
+    '2002-12-25'
     >>> src = '2002/12/25'
     >>> date_format(src)
-    'Dec 25, 2002'
+    '2002-12-25'
     >>> src = 1040814000
     >>> date_format(src)
-    'Dec 25, 2002'
+    '2002-12-25'
     >>> src = '1040814000'
     >>> date_format(src)
-    'Dec 25, 2002'
+    '2002-12-25'
     '''
     return date_cast(date).strftime(format)
 
 
-def warn_until(version_info,
+def warn_until(version,
                message,
                category=DeprecationWarning,
                stacklevel=None,
@@ -1419,12 +1592,13 @@ def warn_until(version_info,
                _dont_call_warnings=False):
     '''
     Helper function to raise a warning, by default, a ``DeprecationWarning``,
-    until the provided ``version_info``, after which, a ``RuntimeError`` will
+    until the provided ``version``, after which, a ``RuntimeError`` will
     be raised to remind the developers to remove the warning because the
     target version has been reached.
 
-    :param version_info: The version info after which the warning becomes a
-                         ``RuntimeError``. For example ``(0, 17)``.
+    :param version: The version info or name after which the warning becomes a
+                    ``RuntimeError``. For example ``(0, 17)`` or ``Hydrogen``
+                    or an instance of :class:`salt.version.SaltStackVersion`.
     :param message: The warning message to be displayed.
     :param category: The warning class to be thrown, by default
                      ``DeprecationWarning``
@@ -1438,10 +1612,17 @@ def warn_until(version_info,
                                 issued. When we're only after the salt version
                                 checks to raise a ``RuntimeError``.
     '''
-    if not isinstance(version_info, tuple):
+    if not isinstance(version, (tuple,
+                                salt._compat.string_types,
+                                salt.version.SaltStackVersion)):
         raise RuntimeError(
-            'The \'version_info\' argument should be passed as a tuple.'
+            'The \'version\' argument should be passed as a tuple, string or '
+            'an instance of \'salt.version.SaltStackVersion\'.'
         )
+    elif isinstance(version, tuple):
+        version = salt.version.SaltStackVersion(*version)
+    elif isinstance(version, salt._compat.string_types):
+        version = salt.version.SaltStackVersion.from_name(version)
 
     if stacklevel is None:
         # Attribute the warning to the calling function, not to warn_until()
@@ -1450,30 +1631,36 @@ def warn_until(version_info,
     if _version_info_ is None:
         _version_info_ = salt.version.__version_info__
 
-    if _version_info_ >= version_info:
+    _version_ = salt.version.SaltStackVersion(*_version_info_)
+
+    if _version_ >= version:
         caller = inspect.getframeinfo(sys._getframe(stacklevel - 1))
         raise RuntimeError(
             'The warning triggered on filename {filename!r}, line number '
             '{lineno}, is supposed to be shown until version '
-            '{until_version!r} is released. Current version is now '
-            '{salt_version!r}. Please remove the warning.'.format(
+            '{until_version} is released. Current version is now '
+            '{salt_version}. Please remove the warning.'.format(
                 filename=caller.filename,
                 lineno=caller.lineno,
-                until_version='.'.join(map(str, version_info)),
-                salt_version='.'.join(map(str, _version_info_))
+                until_version=version.formatted_version,
+                salt_version=_version_.formatted_version
             ),
         )
 
     if _dont_call_warnings is False:
-        warnings.warn(message, category, stacklevel=stacklevel)
+        warnings.warn(
+            message.format(version=version.formatted_version),
+            category,
+            stacklevel=stacklevel
+        )
 
 
 def kwargs_warn_until(kwargs,
-               version_info,
-               category=DeprecationWarning,
-               stacklevel=None,
-               _version_info_=None,
-               _dont_call_warnings=False):
+                      version,
+                      category=DeprecationWarning,
+                      stacklevel=None,
+                      _version_info_=None,
+                      _dont_call_warnings=False):
     '''
     Helper function to raise a warning (by default, a ``DeprecationWarning``)
     when unhandled keyword arguments are passed to function, until the
@@ -1487,8 +1674,9 @@ def kwargs_warn_until(kwargs,
     for the modern strategy for deprecating a function parameter.
 
     :param kwargs: The caller's ``**kwargs`` argument value (a ``dict``).
-    :param version_info: The version info after which the warning becomes a
-                         ``RuntimeError``. For example ``(0, 17)``.
+    :param version: The version info or name after which the warning becomes a
+                    ``RuntimeError``. For example ``(0, 17)`` or ``Hydrogen``
+                    or an instance of :class:`salt.version.SaltStackVersion`.
     :param category: The warning class to be thrown, by default
                      ``DeprecationWarning``
     :param stacklevel: There should be no need to set the value of
@@ -1501,10 +1689,17 @@ def kwargs_warn_until(kwargs,
                                 issued. When we're only after the salt version
                                 checks to raise a ``RuntimeError``.
     '''
-    if not isinstance(version_info, tuple):
+    if not isinstance(version, (tuple,
+                                salt._compat.string_types,
+                                salt.version.SaltStackVersion)):
         raise RuntimeError(
-            'The \'version_info\' argument should be passed as a tuple.'
+            'The \'version\' argument should be passed as a tuple, string or '
+            'an instance of \'salt.version.SaltStackVersion\'.'
         )
+    elif isinstance(version, tuple):
+        version = salt.version.SaltStackVersion(*version)
+    elif isinstance(version, salt._compat.string_types):
+        version = salt.version.SaltStackVersion.from_name(version)
 
     if stacklevel is None:
         # Attribute the warning to the calling function,
@@ -1514,16 +1709,19 @@ def kwargs_warn_until(kwargs,
     if _version_info_ is None:
         _version_info_ = salt.version.__version_info__
 
-    if kwargs or _version_info_ >= version_info:
-        removal_version = '.'.join(str(component) for component in version_info)
-        arg_names = ', '.join('\'{0}\''.format(key) for key in kwargs)
-        warn_until(version_info,
-           message='The following parameter(s) have been deprecated and '
-           'will be removed in {0}: {1}.'.format(removal_version, arg_names),
-           category=category,
-           stacklevel=stacklevel,
-           _version_info_=_version_info_,
-           _dont_call_warnings=_dont_call_warnings
+    _version_ = salt.version.SaltStackVersion(*_version_info_)
+
+    if kwargs or _version_.info >= version.info:
+        arg_names = ', '.join('{0!r}'.format(key) for key in kwargs)
+        warn_until(
+            version,
+            message='The following parameter(s) have been deprecated and '
+                    'will be removed in {0!r}: {1}.'.format(version.string,
+                                                            arg_names),
+            category=category,
+            stacklevel=stacklevel,
+            _version_info_=_version_.info,
+            _dont_call_warnings=_dont_call_warnings
         )
 
 
@@ -1575,10 +1773,32 @@ def compare_versions(ver1='', oper='==', ver2='', cmp_func=None):
         return cmp_result in cmp_map[oper]
 
 
+def compare_dicts(old=None, new=None):
+    '''
+    Compare before and after results from various salt functions, returning a
+    dict describing the changes that were made.
+    '''
+    ret = {}
+    for key in set((new or {}).keys()).union((old or {}).keys()):
+        if key not in old:
+            # New key
+            ret[key] = {'old': '',
+                        'new': new[key]}
+        elif key not in new:
+            # Key removed
+            ret[key] = {'new': '',
+                        'old': old[key]}
+        elif new[key] != old[key]:
+            # Key modified
+            ret[key] = {'old': old[key],
+                        'new': new[key]}
+    return ret
+
+
 def argspec_report(functions, module=''):
     '''
     Pass in a functions dict as it is returned from the loader and return the
-    argspec function sigs
+    argspec function signatures
     '''
     ret = {}
     # TODO: cp.get_file will also match cp.get_file_str. this is the
@@ -1615,10 +1835,10 @@ def memoize(func):
     Deprecation warning wrapper since memoize is now on salt.utils.decorators
     '''
     warn_until(
-        (0, 19),
+        'Helium',
         'The \'memoize\' decorator was moved to \'salt.utils.decorators\', '
         'please start importing it from there. This warning and wrapper '
-        'will be removed on salt > 0.19.0.',
+        'will be removed in Salt {version}.',
         stacklevel=3
 
     )
@@ -1687,7 +1907,7 @@ def is_bin_file(path):
         return None
     try:
         with open(path, 'r') as fp_:
-            return(is_bin_str(fp_.read(2048)))
+            return is_bin_str(fp_.read(2048))
     except os.error:
         return None
 
@@ -1743,3 +1963,99 @@ def repack_dictlist(data):
                 return {}
             ret.update(element)
     return ret
+
+
+def get_group_list(user=None, include_default=True):
+    '''
+    Returns a list of all of the system group names of which the user
+    is a member.
+    '''
+    if HAS_GRP is False or HAS_PWD is False:
+        # We don't work on platforms that don't have grp and pwd
+        # Just return an empty list
+        return []
+    group_names = None
+    if not isinstance(user, string_types):
+        raise Exception
+    if hasattr(os, 'getgrouplist'):
+        # Try os.getgrouplist, available in python >= 3.3
+        log.trace('Trying os.getgrouplist for {0!r}'.format(user))
+        try:
+            group_names = list(os.getgrouplist(user, pwd.getpwnam(user).pw_gid))
+            log.trace('os.getgrouplist for user {0!r}: {1!r}'.format(user, group_names))
+        except Exception:
+            pass
+    else:
+        # Try pysss.getgrouplist
+        log.trace('Trying pysss.getgrouplist for {0!r}'.format(user))
+        try:
+            import pysss
+            group_names = list(pysss.getgrouplist(user))
+            log.trace('pysss.getgrouplist for user {0!r}: {1!r}'.format(user, group_names))
+        except Exception:
+            pass
+    if group_names is None:
+        # Fall back to generic code
+        # Include the user's default group to behave like
+        # os.getgrouplist() and pysss.getgrouplist() do
+        log.trace('Trying generic group list for {0!r}'.format(user))
+        group_names = [g.gr_name for g in grp.getgrall() if user in g.gr_mem]
+        try:
+            default_group = grp.getgrgid(pwd.getpwnam(user).pw_gid).gr_name
+            if default_group not in group_names:
+                group_names.append(default_group)
+        except KeyError:
+            # If for some reason the user does not have a default group
+            pass
+        log.trace('Generic group list for user {0!r}: {1!r}'.format(user, group_names))
+    if include_default is False:
+        # Historically, saltstack code for getting group lists did not
+        # include the default group. Some things may only want
+        # supplemental groups, so include_default=False omits the users
+        # default group.
+        try:
+            default_group = grp.getgrgid(pwd.getpwnam(user).pw_gid).gr_name
+            group_names.remove(default_group)
+        except KeyError:
+            # If for some reason the user does not have a default group
+            pass
+    return sorted(set(group_names))
+
+
+def get_group_dict(user=None, include_default=True):
+    '''
+    Returns a dict of all of the system groups as keys, and group ids
+    as values, of which the user is a member.
+    E.g: {'staff': 501, 'sudo': 27}
+    '''
+    if HAS_GRP is False or HAS_PWD is False:
+        # We don't work on platforms that don't have grp and pwd
+        # Just return an empty dict
+        return {}
+    group_dict = {}
+    group_names = get_group_list(user, include_default=include_default)
+    for group in group_names:
+        group_dict.update({group: grp.getgrnam(group).gr_gid})
+    return group_dict
+
+
+def get_gid_list(user=None, include_default=True):
+    '''
+    Returns a list of all of the system group IDs of which the user
+    is a member.
+    '''
+    if HAS_GRP is False or HAS_PWD is False:
+        # We don't work on platforms that don't have grp and pwd
+        # Just return an empty list
+        return []
+    gid_list = [gid for (group, gid) in salt.utils.get_group_dict(user, include_default=include_default).items()]
+    return sorted(set(gid_list))
+
+
+def total_seconds(td):
+    '''
+    Takes a timedelta and returns the total number of seconds
+    represented by the object. Wrapper for the total_seconds()
+    method which does not exist in versions of Python < 2.7.
+    '''
+    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
